from datetime import timedelta

from django.conf import settings
from django.utils.module_loading import import_string

USER_SERIALIZER = getattr(settings, 'USER_SERIALIZER', None)
if USER_SERIALIZER:
    USER_SERIALIZER = import_string(USER_SERIALIZER)

AUTH_TOKEN_VALIDITY = getattr(settings, 'AUTH_TOKEN_VALIDITY', timedelta(days=1))

REGISTRATION_ENABLED = getattr(settings, 'REGISTRATION_ENABLED', False)
REGISTRATION_EMAIL_CONFIRM = getattr(settings, 'REGISTRATION_EMAIL_CONFIRM', False)
REGISTRATION_EMAIL_CONFIRM_MODEL_FIELD = getattr(
    settings, 'REGISTRATION_EMAIL_CONFIRM_MODEL_FIELD', False)
REGISTRATION_EMAIL_CONFIRM_TOKEN_VALIDITY = getattr(
    settings, 'REGISTRATION_EMAIL_CONFIRM_TOKEN_VALIDITY', timedelta(days=1))
REGISTRATION_EMAIL = getattr(
    settings, 'REGISTRATION_EMAIL', None)
REGISTRATION_CONFIRM_REDIRECT_PATH = getattr(
    settings, 'REGISTRATION_CONFIRM_REDIRECT_PATH', '/')
REGISTRATION_CONFIRM_INVALID_REDIRECT_PATH = getattr(
    settings, 'REGISTRATION_CONFIRM_INVALID_REDIRECT_PATH', None)



from base64 import urlsafe_b64encode

from django.contrib.auth import get_user_model
from django.core.mail import send_mail
from django.urls import reverse
from rest_framework.exceptions import ValidationError

from .models import EmailConfirmationToken
from .settings import REGISTRATION_EMAIL, \
    REGISTRATION_EMAIL_CONFIRM_MODEL_FIELD


def send_confirmation_email(user: get_user_model()):
    """
    Send a confirmation email to the specified user.
    May only be called for a user object with an unverified email address.
    """
    assert not getattr(
        user, REGISTRATION_EMAIL_CONFIRM_MODEL_FIELD), 'email already confirmed'

    token = urlsafe_b64encode(
        EmailConfirmationToken.create_token_for_user(user)).decode()

    path = reverse('rest_authtoken:confirm_email', args=(token,))
    url = REGISTRATION_EMAIL['BASE_URL'] + path

    message = REGISTRATION_EMAIL['MESSAGE'].format(
        username=user.username,
        url=url,
    )

    send_mail(
        REGISTRATION_EMAIL['SUBJECT'],
        message,
        REGISTRATION_EMAIL['FROM'],
        [user.email],
        fail_silently=False
    )



from base64 import urlsafe_b64decode, urlsafe_b64encode

from django.contrib.auth import authenticate, user_logged_in
from django.db import transaction
from django.http import HttpResponseBadRequest, HttpResponseRedirect
from rest_framework import status
from rest_framework.exceptions import ValidationError
from rest_framework.permissions import IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework.viewsets import GenericViewSet, ViewSet

from .models import AuthToken, EmailConfirmationToken
from .serializers import UserRegistrationSerializer
from .settings import REGISTRATION_CONFIRM_INVALID_REDIRECT_PATH, \
    REGISTRATION_CONFIRM_REDIRECT_PATH, \
    REGISTRATION_EMAIL_CONFIRM_MODEL_FIELD, USER_SERIALIZER


class LoginViewSet(ViewSet):
    @staticmethod
    def create(request: Request) -> Response:
        user = authenticate(
            username=request.data.get('username'),
            password=request.data.get('password'))
        if not user:
            return Response(
                'invalid credentials',
                status=status.HTTP_401_UNAUTHORIZED)

        token = AuthToken.create_token_for_user(user)

        data = {
            'token': urlsafe_b64encode(token),
        }

        if USER_SERIALIZER:
            data['user'] = USER_SERIALIZER(
                instance=user, read_only=True).data

        user_logged_in.send(sender=user.__class__, request=request, user=user)

        return Response(data)


class LogoutView(APIView):
    permission_classes = IsAuthenticated,

    @staticmethod
    def delete(request: Request) -> Response:
        auth_header = request.META.get('HTTP_AUTHORIZATION').split()

        if len(auth_header) != 2 or auth_header[0].lower() != 'token':
            raise ValidationError('no token')

        try:
            auth_token = urlsafe_b64decode(auth_header[1])
        except ValueError:
            raise ValidationError('invalid token')

        auth_token = AuthToken.get_token(auth_token)

        if not auth_token:
            raise ValidationError('invalid token')

        auth_token.delete()

        return Response(status=status.HTTP_204_NO_CONTENT)


class RegisterViewSet(GenericViewSet):
    serializer_class = UserRegistrationSerializer

    def create(self, request: Request) -> Response:
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return Response({'success': True}, status=status.HTTP_201_CREATED)


@transaction.atomic
def confirm_email(request, token: str):
    def invalid_action():
        if REGISTRATION_CONFIRM_INVALID_REDIRECT_PATH:
            return HttpResponseRedirect(REGISTRATION_CONFIRM_INVALID_REDIRECT_PATH)
        else:
            return HttpResponseBadRequest('invalid token!')

    try:
        token = urlsafe_b64decode(token)
    except ValueError:
        return invalid_action()
    token = EmailConfirmationToken.get_token(token)
    if not token:
        return invalid_action()

    user = token.user
    if user.email != token.email:
        return invalid_action()

    setattr(user, REGISTRATION_EMAIL_CONFIRM_MODEL_FIELD, True)
    user.save(update_fields=(REGISTRATION_EMAIL_CONFIRM_MODEL_FIELD,))
    token.delete()

    return HttpResponseRedirect(REGISTRATION_CONFIRM_REDIRECT_PATH)



from datetime import timedelta
from hashlib import sha512
from os import urandom
from typing import Optional, Union

from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import models
from django.utils import timezone

from .settings import AUTH_TOKEN_VALIDITY, REGISTRATION_EMAIL_CONFIRM_TOKEN_VALIDITY


class AbstractToken(models.Model):
    class Meta:
        abstract = True

    TOKEN_VALIDITY = AUTH_TOKEN_VALIDITY

    hashed_token = models.BinaryField(primary_key=True)

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name='%(class)ss',
        on_delete=models.CASCADE)

    created = models.DateTimeField(default=timezone.now)

    def __str__(self) -> str:
        return '{}: {}'.format(self.user, self.hashed_token)

    @property
    def age(self) -> timedelta:
        return timezone.now() - self.created

    def logout(self, token: Union[str, None] = None):
        """
        Log this token out.
        """
        self.delete()

    @classmethod
    def create_token_for_user(cls, user: get_user_model(), **kwargs) -> bytes:
        """
        Create a new random auth token for user.
        """
        token = urandom(48)
        cls.objects.create(
            hashed_token=cls._hash_token(token),
            user=user,
            **kwargs)
        return token

    @classmethod
    def get_user_for_token(cls, token: bytes) -> Optional[get_user_model()]:
        auth_token = cls.get_token(token)
        if auth_token:
            return auth_token.user

    @classmethod
    def get_token(cls, token: bytes) -> Optional['AbstractToken']:
        try:
            auth_token = cls.objects.select_related('user').get(
                hashed_token=cls._hash_token(token))

            if auth_token.age > cls.TOKEN_VALIDITY:
                # token expired.
                auth_token.delete()
                return None

            return auth_token
        except cls.DoesNotExist:
            return None

    @classmethod
    def clear_expired_tokens(cls) -> int:
        """
        Clear tokens that are expired.
        """
        valid_min_creation = timezone.now() - cls.TOKEN_VALIDITY
        deleted_count, detail = cls.objects.filter(created__lt=valid_min_creation).delete()
        return deleted_count

    @staticmethod
    def _hash_token(token: bytes) -> bytes:
        """
        Hash a token.
        """
        return sha512(token).digest()


class AuthToken(AbstractToken):
    pass


class EmailConfirmationToken(AbstractToken):
    TOKEN_VALIDITY = REGISTRATION_EMAIL_CONFIRM_TOKEN_VALIDITY

    email = models.EmailField()

    @classmethod
    def create_token_for_user(cls, user: get_user_model()) -> bytes:
        return super().create_token_for_user(user, email=user.email)



from base64 import urlsafe_b64decode

from rest_framework.authentication import BaseAuthentication, get_authorization_header
from rest_framework.exceptions import AuthenticationFailed

from .models import AuthToken


class AuthTokenAuthentication(BaseAuthentication):
    def authenticate(self, request):
        """
        Authenticate the request and return a two-tuple of (user, token).
        """
        auth = get_authorization_header(request).split()

        if not auth or auth[0].lower() != b'token':
            return None

        if len(auth) == 1:
            msg = 'invalid auth header, No credentials provided.'
            raise AuthenticationFailed(msg)
        elif len(auth) > 2:
            msg = 'invalid auth token'
            raise AuthenticationFailed(msg)

        try:
            token = urlsafe_b64decode(auth[1])
        except ValueError:
            msg = 'invalid auth token'
            raise AuthenticationFailed(msg)

        return self.authenticate_credentials(token, request)

    def authenticate_credentials(self, token: bytes, request=None):
        """
        Authenticate the token with optional request for context.
        """
        user = AuthToken.get_user_for_token(token)

        if user is None:
            raise AuthenticationFailed('invalid auth token')

        if not user.is_active:
            raise AuthenticationFailed('invalid auth token')

        return user, token

    def authenticate_header(self, request):
        return 'Token'



# Generated by Django 3.1 on 2020-08-22 13:20

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('rest_authtoken', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='authtoken',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='authtokens', to=settings.AUTH_USER_MODEL),
        ),
        migrations.CreateModel(
            name='EmailConfirmationToken',
            fields=[
                ('hashed_token', models.BinaryField(primary_key=True, serialize=False)),
                ('created', models.DateTimeField(default=django.utils.timezone.now)),
                ('email', models.EmailField(max_length=254)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='emailconfirmationtokens', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



# Generated by Django 2.0.7 on 2018-07-19 11:46

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='AuthToken',
            fields=[
                ('hashed_token', models.BinaryField(primary_key=True, serialize=False)),
                ('created', models.DateTimeField(default=django.utils.timezone.now)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='auth_tokens', to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



from django.core.management import BaseCommand
from rest_authtoken.models import AuthToken, EmailConfirmationToken


class Command(BaseCommand):
    help = 'Delete all epxpired auth and email confirmation tokens'

    def handle(self, *args, **options):
        deleted_auth_tokens = AuthToken.clear_expired_tokens()
        self.stdout.write(
            self.style.SUCCESS(
                'Successfully cleared {} auth tokens'.format(deleted_auth_tokens)))
        deleted_email_tokens = EmailConfirmationToken.clear_expired_tokens()
        self.stdout.write(
            self.style.SUCCESS(
                'Successfully cleared {} email confirmation tokens'.format(
                    deleted_email_tokens)))



from django.urls import path
from rest_framework.routers import SimpleRouter

from . import views
from .settings import REGISTRATION_EMAIL_CONFIRM, REGISTRATION_ENABLED

app_name = 'rest_authtoken'

router = SimpleRouter()
router.register('login', views.LoginViewSet, basename='login')

if REGISTRATION_ENABLED:
    router.register('register', views.RegisterViewSet, basename='register')

urlpatterns = router.urls

urlpatterns += [
    path('logout/', views.LogoutView.as_view()),
]

if REGISTRATION_ENABLED and REGISTRATION_EMAIL_CONFIRM:
    urlpatterns += [
        path('register/confirm/<str:token>/', views.confirm_email,
             name='confirm_email'),
    ]



import setuptools

with open('README.md') as fh:
    long_description = fh.read()

setuptools.setup(
    name='django-rest-authtoken',
    version='2.1.4',
    author='Pascal Wichmann',
    author_email='wichmannpas@gmail.com',
    description='A simple token-based auth backend for Django Rest Framework storing cryptographically hashed tokens on server-side.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/wichmannpas/django-rest-authtoken',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
)



from smtplib import SMTPException

from django.contrib.auth import get_user_model
from django.db import transaction
from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from .email_confirmation import send_confirmation_email
from .settings import REGISTRATION_EMAIL_CONFIRM, REGISTRATION_EMAIL_CONFIRM_MODEL_FIELD


class UserRegistrationSerializer(serializers.ModelSerializer):
    class Meta:
        model = get_user_model()
        fields = [
            'username',
            'password',
        ]
        if REGISTRATION_EMAIL_CONFIRM:
            fields += [
                'email',
            ]
            extra_kwargs = {
                'email': {
                    'required': True,
                },
            }

    @staticmethod
    def validate_username(value):
        """
        Make sure that no user with this natural key exists.
        This can be used to make usernames case-insensitive, e.g.,
        disallow the registration of two different usernames which
        are equal but with different case.
        """
        try:
            user = get_user_model().objects.get_by_natural_key(value)
        except get_user_model().DoesNotExist:
            return value
        else:
            raise ValidationError('already exists')

    @staticmethod
    def validate_password(value):
        if len(value) < 6:
            raise ValidationError('password too short')
        return value

    @transaction.atomic
    def create(self, validated_data):
        password = validated_data.pop('password')

        instance = self.Meta.model(**validated_data)
        instance.set_password(password)

        if REGISTRATION_EMAIL_CONFIRM:
            setattr(instance, REGISTRATION_EMAIL_CONFIRM_MODEL_FIELD, False)

        instance.save()

        if REGISTRATION_EMAIL_CONFIRM:
            try:
                send_confirmation_email(instance)
            except SMTPException:
                raise ValidationError('failed to send confirmation email')

        return instance
